import abc
import datetime
import croniter
import inspect
import random

from .utils import ObjectMetaclass

class _Trigger(metaclass=ObjectMetaclass):

    @abc.abstractmethod
    def check(self):
        pass

class AQITrigger(_Trigger):

    def __init__(self, check_func, aqi_sensor):
        self.check_func = check_func
        self.aqi_sensor = aqi_sensor

    def check(self):
        aqi = self.aqi_sensor.get_aqi()
        return self.check_func(aqi)

class IsoWeekdayTrigger(_Trigger):

    # TODO: test timezone

    def __init__(self, isoweekdays, time_sensor):
        self.isoweekdays = isoweekdays
        self.time_sensor = time_sensor

    def check(self):
        now = self.time_sensor.get_current_datetime()
        return now.isoweekday() in self.isoweekdays

class TimeTrigger(_Trigger):

    # TODO: test timezone

    def __init__(self, times, time_sensor):
        self.times = times
        self.time_sensor = time_sensor

    def check(self):
        now = self.time_sensor.get_current_datetime()
        now_minutes = 60*now.hour + now.minute
        for minutes in self.times:
            if minutes == now_minutes:
                return True
        return False

class DateTrigger(_Trigger):

    def __init__(self, dates, time_sensor):
        self.dates = dates
        self.time_sensor = time_sensor

    def check(self):
        today = self.time_sensor.get_current_datetime().date()
        for date in self.dates:
            if date == today:
                return True
        return False

class CronTrigger(_Trigger):

    def __init__(self, cron, time_sensor):
        self.cron = cron
        self.time_sensor = time_sensor

    def check(self):
        now = self.time_sensor.get_current_datetime()
        return croniter.croniter.match(self.cron, now)

class RandomTrigger(_Trigger):

    def __init__(self, probability):
        self.probability = probability

    def check(self):
        return random.random() < self.probability

class _SunTrigger(_Trigger):

    def __init__(self, timedeltas, time_sensor, sun_sensor):
        self.timedeltas = timedeltas
        self.time_sensor = time_sensor
        self.sun_sensor = sun_sensor
        self.sun_sensor_method = getattr(self.sun_sensor,
                self.sun_sensor_method_name)

    def check(self):
        sun_time = self.sun_sensor_method()
        sun_minutes = 60*sun_time.hour + sun_time.minute
        now = self.time_sensor.get_current_datetime()
        now_minutes = 60*now.hour + now.minute
        return now_minutes - sun_minutes in self.timedeltas

class SunriseTrigger(_SunTrigger):

    sun_sensor_method_name = 'get_sunrise'

class SunsetTrigger(_SunTrigger):

    sun_sensor_method_name = 'get_sunset'

class TemperatureTrigger(_Trigger):

    def __init__(self, check_func, weather_sensor):
        self.check_func = check_func
        self.weather_sensor = weather_sensor

    def check(self):
        temp = self.weather_sensor.current_temperature()
        return self.check_func(temp)

class RadonTrigger(_Trigger):

    def __init__(self, check_func, radon_sensor):
        self.check_func = check_func
        self.radon_sensor = radon_sensor

    def check(self):
        radon = self.radon_sensor.current_radon()
        return self.check_func(radon)

class WebhookTrigger(_Trigger):

    # TODO: test webhook trigger

    def __init__(self, path, webhook_sensor):
        self.path = path
        self.webhook_sensor = webhook_sensor

    def check(self):
        return self.webhook_sensor.path == self.path and \
                self.webhook_sensor.method == 'POST'



import pyflowater

from .base import Provider, Device

class MoenProvider(Provider):

    def __init__(self, username, password):
        self.flo = pyflowater.PyFlo(username, password)

    def get_device(self, device_id, device_name, device_description):
        location_id = self._get_location_id(device_id)
        device = MoenDevice.API(self.flo, location_id, device_id)
        return MoenDevice(device, device_name, device_description)

    def _get_location_id(self, device_id):
        for location in self.flo.locations():
            for device in location['devices']:
                if device['id'] == device_id:
                    return location['id']

class MoenDevice(Device):

    def turn_on(self):
        self.device.open_valve()

    def turn_off(self):
        self.device.close_valve()

    def set_mode(self, mode, params):
        return self.device.set_mode(mode, params)

    class API(object):

        def __init__(self, flo, location_id, device_id):
            self.flo = flo
            self.location_id = location_id
            self.device_id = device_id

        def open_valve(self):
            self.flo.open_valve(self.device_id)

        def close_valve(self):
            self.flo.close_valve(self.device_id)

        def set_mode(self, mode, params):
            return self.flo.set_mode(self.location_id, mode, additional_params=params)



import logging

from .exceptions import PyDomoticConfigParsingError
from .sensors import (TimeSensor, WebhookSensor, DeviceSensor, WeatherSensor,
        AQISensor, SunSensor)

logger = logging.getLogger(__name__)

class Context(object):

    def __init__(self, latitude, longitude, aqi_api_key, weather_api_key,
                weather_data_cache_seconds, timezone):
        self._aqi_api_key = aqi_api_key
        self._latitude = latitude
        self._longitude = longitude
        self._timezone = timezone
        self._weather_api_key = weather_api_key
        self._weather_data_cache_seconds = weather_data_cache_seconds

        self._device_sensors = {}
        self._aqi_sensor = None
        self._sun_sensor = None
        self._time_sensor = None
        self._weather_sensor = None
        self._webhook_sensor = None

        self._context = None
        self._sensors = None
        self.devices = {}

    @staticmethod
    def from_yaml(triggers):
        from .parsers import _parse_string
        try:
            location = triggers.get('location', {})
            latitude = location.get('latitude')
            longitude = location.get('longitude')
        except Exception as e:
            logger.debug('failed to parse location, ignoring: '
                    f'[{e.__class__.__name__}] {e}')
            latitude = longitude = None

        try:
            timezone = triggers.get('timezone')
        except Exception as e:
            logger.debug('failed to parse timezone, ignoring: '
                    f'[{e.__class__.__name__}] {e}')
            timezone = None

        try:
            aqi = triggers.get('aqi', {})
            aqi_api_key = _parse_string(aqi.get('api_key'))
        except Exception as e:
            logger.debug('failed to parse aqi api_key, ignoring: '
                    f'[{e.__class__.__name__}] {e}')
            aqi_api_key = None

        try:
            weather = triggers.get('weather', {})
            weather_api_key = _parse_string(weather.get('api_key'))
        except Exception as e:
            logger.debug('failed to parse weather api_key, ignoring: '
                    f'[{e.__class__.__name__}] {e}')
            weather_api_key = None

        try:
            weather = triggers.get('weather', {})
            weather_data_cache_seconds = weather.get('data_cache_seconds')
        except Exception as e:
            logger.debug('failed to parse weather data_cache_seconds, ignoring: '
                    f'[{e.__class__.__name__}] {e}')
            weather_data_cache_seconds = None

        return Context(
                latitude, longitude, aqi_api_key, weather_api_key,
                weather_data_cache_seconds, timezone)

    @property
    def latitude(self):
        if self._latitude is None:
            raise PyDomoticConfigParsingError(
                    'latitude value required for location')
        if not isinstance(self._latitude, (int, float)):
            raise PyDomoticConfigParsingError(
                    'latitude must be a number, not '
                    f'{self._latitude.__class__.__name__}')
        return self._latitude

    @property
    def longitude(self):
        if self._longitude is None:
            raise PyDomoticConfigParsingError(
                    'longitude value required for location')
        if not isinstance(self._longitude, (int, float)):
            raise PyDomoticConfigParsingError(
                    'longitude must be a number, not '
                    f'{self._longitude.__class__.__name__}')
        return self._longitude

    @property
    def timezone(self):
        if self._timezone is None:
            raise PyDomoticConfigParsingError('timezone is required')
        if not isinstance(self._timezone, str):
            raise PyDomoticConfigParsingError(
                    'timezone must be a string, not '
                    f'{self._timezone.__class__.__name__}')
        return self._timezone

    @property
    def aqi_api_key(self):
        if not self._aqi_api_key:
            raise PyDomoticConfigParsingError('aqi api key required')
        if not isinstance(self._aqi_api_key, str):
            raise PyDomoticConfigParsingError(
                    'aqi api_key must be a string, not '
                    f'{self._aqi_api_key.__class__.__name__}')
        return self._aqi_api_key

    @property
    def weather_api_key(self):
        if not self._weather_api_key:
            raise PyDomoticConfigParsingError('weather api key required')
        if not isinstance(self._weather_api_key, str):
            raise PyDomoticConfigParsingError(
                    'weather api_key must be a string, not '
                    f'{self._aqi_api_key.__class__.__name__}')
        return self._weather_api_key

    @property
    def weather_data_cache_seconds(self):
        if self._weather_data_cache_seconds is None:
            return None
        if not isinstance(self._weather_data_cache_seconds, int):
            raise PyDomoticConfigParsingError(
                    'weather data_cache_seconds must be an integer, not '
                    f'{self._weather_data_cache_seconds.__class__.__name__}')
        return self._weather_data_cache_seconds

    @property
    def aqi_sensor(self):
        if self._aqi_sensor is None:
            self._aqi_sensor = AQISensor(
                    self.aqi_api_key, self.latitude, self.longitude)
        return self._aqi_sensor

    @property
    def sun_sensor(self):
        if self._sun_sensor  is None:
            self._sun_sensor = SunSensor(
                    self.latitude, self.longitude, self.time_sensor)
        return self._sun_sensor

    @property
    def time_sensor(self):
        if self._time_sensor is None:
            if self._timezone is not None:
                self._time_sensor = TimeSensor(timezone=self.timezone)
            elif self._longitude is not None and self._latitude is not None:
                self._time_sensor = TimeSensor(latitude=self.latitude,
                        longitude=self.longitude)
            else:
                raise PyDomoticConfigParsingError(
                        'either timezone or latitude/longitude required')
        return self._time_sensor

    @property
    def weather_sensor(self):
        if self._weather_sensor is None:
            self._weather_sensor = WeatherSensor(
                    self.weather_api_key, self.latitude, self.longitude,
                    data_cache_seconds=self.weather_data_cache_seconds)
        return self._weather_sensor

    @property
    def webhook_sensor(self):
        if self._webhook_sensor is None:
            self._webhook_sensor = WebhookSensor()
        return self._webhook_sensor

    def device_sensor(self, name):
        device = self.devices.get(name, None)
        if device is None:
            raise PyDomoticConfigParsingError(f'sensor device {name} not found')
        if not self._device_sensors.get(device.name):
            self._device_sensors[device.name] = DeviceSensor(device)
        return self._device_sensors[device.name]

    @property
    def sensors(self):
        if self._sensors is None:
            self._sensors = {}
            for sensor in ('aqi_sensor', 'sun_sensor', 'time_sensor',
                    'weather_sensor'):
                try:
                    self._sensors[sensor] = getattr(self, sensor)
                except Exception as e:
                    logger.debug(f'{sensor} not configured, ignoring: '
                            f'[{e.__class__.__name__}] {e}')
        return self._sensors

    @property
    def context(self):
        if not self._context:
            # TODO: make a copy to to pass to action.run?
            self._context = {
                    'devices': self.devices,
                    'sensors': self.sensors,
            }
        return self._context



import logging

import pyfujitseu.api
import pyfujitseu.splitAC

from .base import Provider, Device

logger = logging.getLogger(__name__)

# see https://docs.aylanetworks.com/reference/getting_started

class FujitsuProvider(Provider):

    def __init__(self, username, password, tokenpath='/tmp/token.txt'):
        self.fujitsu = pyfujitseu.api.Api(username, password, tokenpath=tokenpath)

    def get_device(self, device_id, device_name, device_description):
        device = pyfujitseu.splitAC.splitAC(dsn=device_id, api=self.fujitsu)
        return FujitsuDevice(device, device_name, device_description)

class FujitsuDevice(Device):

    def turn_on(self):
        self.device.turnOn()
        logger.debug('device "%s" turned on', self.name)

    def turn_off(self):
        self.device.turnOff()
        logger.debug('device "%s" turned off', self.name)

    def switch(self):
        logger.debug('switch action not implemented for fujitsu device')



import astral
import astral.sun
import datetime
import requests
import zoneinfo

from .utils import cache_value, ObjectMetaclass

class _Sensor(metaclass=ObjectMetaclass):
    pass

class AQISensor(_Sensor):

    aqi_url = 'https://www.airnowapi.org/aq/observation/latLong/current'
    timeout = 5 # seconds

    def __init__(self, api_key, latitude, longitude):
        self.location = (latitude, longitude)
        self.params = {
                'latitude': latitude,
                'longitude': longitude,
                'api_key': api_key,
                'format': 'application/json',
        }

    @cache_value(minutes=15, fallback_on_error=True)
    def get_aqi(self):
        try:
            resp = requests.get(self.aqi_url, params=self.params, timeout=self.timeout)
            resp.raise_for_status()
            data = resp.json()
        except Exception as e:
            raise AQISensorError(
                    f'error getting current aqi: [{e.__class__.__name__}] {e}')
        aqi = None
        try:
            for d in data:
                if d['ParameterName'] != 'PM2.5':
                    continue
                aqi = d['AQI']
                break
        except Exception as e:
            raise AQISensorError(
                    f'malformed response json [{e.__class__.__name__}] {e}')
        if aqi is None:
            raise AQISensorError(
                    f'unable to find aqi for location {self.location}')
        return aqi

class AQISensorError(Exception):
    pass

class SunSensor(_Sensor):

    # TODO: test timezone

    def __init__(self, latitude, longitude, time_sensor):
        self.observer = astral.Observer(latitude=latitude, longitude=longitude)
        self.time_sensor = time_sensor

    @cache_value(hours=12)
    def get_sunrise(self):
        now = self.time_sensor.get_current_datetime()
        return astral.sun.sunrise(self.observer, date=now,
                tzinfo=self.time_sensor.tzinfo)

    @cache_value(hours=12)
    def get_sunset(self):
        now = self.time_sensor.get_current_datetime()
        return astral.sun.sunset(self.observer, date=now,
                tzinfo=self.time_sensor.tzinfo)

class TimeSensor(_Sensor):

    # TODO: test

    def __init__(self, latitude=None, longitude=None, timezone=None):
        if latitude is None and longitude is None and timezone is None:
            raise TypeError('timezone or latitude/longitude required')
        self.latitude = latitude
        self.longitude = longitude
        self.timezone = timezone
        self._tzinfo = None

    def _get_timezone(self):
        if self.timezone:
            return self.timezone
        import timezonefinder
        return timezonefinder.TimezoneFinder().timezone_at(
                lng=self.longitude, lat=self.latitude)

    def get_current_datetime(self):
        return datetime.datetime.now(tz=self.tzinfo)

    @property
    def tzinfo(self):
        # TODO: test
        if self._tzinfo is None:
            if self.timezone or (
                    self.latitude is not None and self.longitude is not None):
                self._tzinfo = zoneinfo.ZoneInfo(self._get_timezone())
            else:
                self._tzinfo = datetime.timezone.utc
        return self._tzinfo

class WeatherSensor(_Sensor):

    def __init__(self, api_key, latitude, longitude, data_cache_seconds=None):
        import pyowm
        self.owm_mgr = pyowm.OWM(api_key).weather_manager()
        self.location = (latitude, longitude)

        if data_cache_seconds is not None:
            self._weather = cache_value(seconds=data_cache_seconds)(self._weather)

    def _weather(self):
        return self.owm_mgr.weather_at_coords(*self.location).weather

    def current_temperature(self):
        # returns float with 2 decimal points
        return self._weather().temperature('fahrenheit').get('temp')

    def current_humidity(self):
        # returns percentage value between 0 and 100
        return self._weather().humidity

    def current_pressure(self):
        return self._weather().barometric_pressure(unit='inHg').get('press')

class WebhookSensor(_Sensor):

    def __init__(self):
        self.path = None
        self.method = None

    def set_webhook_request(self, event):
        # TODO: test setting request from various event types
        http = event.get('requestContext', {}).get('http', {})
        self.path = http.get('path')
        self.method = http.get('method')

class DeviceSensor(_Sensor):

    def __init__(self, device):
        self.device = device

    def __getattr__(self, attr):
        if attr == 'device':
            return self.device
        if attr == '_name':
            return 'device_sensor'
        if attr in dir(self.device):
            return getattr(self.device, attr)
        return super().__getattr__(self, attr)

    @property
    def name(self):
        return f'{super().name} {self.device.name}'



import logging

logger = logging.getLogger(__name__)

class Component(object):

    def __init__(self, name, ifs, thens, elses, enabled=True):
        self.name = name or 'unknown'
        self.ifs = ifs
        self.thens = thens
        self.elses = elses
        self.enabled = enabled

    def run(self):
        logger.debug('running component %s', self.name)
        checked = True
        for trigger in self.ifs:
            logger.debug('checking trigger %s', trigger.name)
            if not trigger.check():
                logger.debug('trigger failed')
                checked = False
                break
            logger.debug('trigger passes')

        if checked:
            logger.debug('all triggers passed')
            exception = self._run_actions(self.thens)
        else:
            exception = self._run_actions(self.elses)

        if exception:
            raise exception

    def _run_actions(self, actions):
        exception = None
        for action in actions:
            logger.debug('running action %s', action.name)
            try:
                action.run()
            except Exception as e:
                logger.exception(f'failure running action {action.name}')
                exception = e
        return exception



import abc

from .utils import ObjectMetaclass, import_method

class _Action(metaclass=ObjectMetaclass):

    @abc.abstractmethod
    def run(self):
        pass

class _DeviceAction(_Action):

    required_class_attrs = ['device_action_method_name']

    def __init__(self, device):
        self.device = device

    def run(self):
        getattr(self.device, self.device_action_method_name)()

    @property
    def name(self):
        return f'{super().name} {self.device.name}'

class TurnOnAction(_DeviceAction):

    device_action_method_name = 'turn_on'

class TurnOffAction(_DeviceAction):

    device_action_method_name = 'turn_off'

class SwitchAction(_DeviceAction):

    device_action_method_name = 'switch'

class SetModeAction(_Action):

    def __init__(self, device, mode, extra_params):
        self.device = device
        self.mode = mode
        self.extra_params = extra_params

    def run(self):
        self.device.set_mode(self.mode, self.extra_params)

    @property
    def name(self):
        return f'{super().name} {self.device.name} {self.mode}'

class ExecuteCodeAction(_Action):

    def __init__(self, import_path, context):
        self.import_path = import_path
        self.context = context
        self._execute_method = None

    @property
    def execute_method(self):
        if not self._execute_method:
            self._execute_method = import_method(self.import_path)
        return self._execute_method

    def run(self):
        self.execute_method(self.context)

    @property
    def name(self):
        return f'{super().name} {self.import_path}'



import requests

from .base import Provider, Device
from ..utils import cache_value

class AirthingsAPI(object):

    _auth_token_url = 'https://accounts-api.airthings.com/v1/token'
    _auth_token_data = {
            'grant_type': 'client_credentials',
            'scope': 'read:device:current_values',
    }
    _auth_headers = {}
    _samples_url = 'https://ext-api.airthings.com/v1/devices/{}/latest-samples'

    def __init__(self, client_id, client_secret, data_cache_seconds=None, timeout=None):
        self._auth_credentials = (client_id, client_secret)
        self._timeout = timeout
        if data_cache_seconds:
            self.fetch_data = cache_value(seconds=data_cache_seconds)(self.fetch_data)

    def _get_auth_headers(self):
        resp = requests.post(
                self._auth_token_url,
                data=self._auth_token_data,
                auth=self._auth_credentials,
                timeout=self._timeout,
        )
        resp.raise_for_status()
        self._auth_headers['Authorization'] = f'Bearer {resp.json().get("access_token")}'
        return self._auth_headers

    def fetch_data(self, device_id):
        resp = requests.get(
                url=self._samples_url.format(device_id),
                headers=self._get_auth_headers(),
                timeout=self._timeout,
        )
        resp.raise_for_status()
        return resp.json().get('data')

    def get_device(self, device_id):
        return self.device(device_id, self)

    class device(object):

        def __init__(self, device_id, api):
            self.device_id = device_id
            self.api = api

        def fetch_data(self):
            return self.api.fetch_data(self.device_id)

        def get_radon(self):
            return self.fetch_data()['radonShortTermAvg'] / 37  # convert Bq/m3 to pCi/L

        def get_temperature(self):
            return self.fetch_data()['temp'] * 1.8 + 32  # convert C to F

        def get_humidity(self):
            return self.fetch_data()['humidity']

        def get_battery(self):
            return self.fetch_data()['battery']

class AirthingsProvider(Provider):

    def __init__(self, client_id, client_secret, data_cache_seconds=None, timeout=None):
        self.api = AirthingsAPI(client_id, client_secret,
                data_cache_seconds=data_cache_seconds, timeout=timeout)

    def get_device(self, device_id, device_name, device_description):
        device = self.api.get_device(device_id)
        return AirthingsDevice(device, device_name, device_description)

class AirthingsDevice(Device):

    def current_radon(self):
        return self.device.get_radon()

    def current_temperature(self):
        return self.device.get_temperature()

    def current_humidity(self):
        return self.device.get_humidity()



import datetime
import croniter
import logging
import os
import re
import yaml

from .actions import (TurnOnAction, TurnOffAction, SwitchAction, SetModeAction,
        ExecuteCodeAction)
from .components import Component
from .context import Context
from .exceptions import PyDomoticConfigParsingError
from .providers.noop import NoopProvider
from .triggers import (AQITrigger, TimeTrigger, IsoWeekdayTrigger, DateTrigger,
        CronTrigger, RandomTrigger, SunriseTrigger, SunsetTrigger,
        TemperatureTrigger, RadonTrigger, WebhookTrigger)

logger = logging.getLogger(__name__)

def parse_yaml(config_file=None, s3=None):
    reader = _get_config_reader(config_file, s3)
    data = reader.read() if reader else None
    return parse_raw_yaml(data)

def parse_raw_yaml(raw_conf):
    conf = yaml.safe_load(raw_conf) if raw_conf else {}
    context = Context.from_yaml(conf.get('triggers', {}))
    context.providers = _parse_providers(conf.get('providers', {}))
    context.devices = _parse_devices(conf.get('devices', {}), context.providers)
    components = _parse_components(conf.get('automations', {}), context)
    return components, context

def _get_config_reader(config_file, s3):
    conf_env = os.environ.get('PYDOMOTIC_CONFIG_FILE')
    s3_env = os.environ.get('PYDOMOTIC_CONFIG_S3')
    logger.debug(f'found environment variables: '
            f'PYDOMOTIC_CONFIG_FILE={conf_env}, PYDOMOTIC_CONFIG_S3={s3_env}')
    if config_file:
        logger.debug(f'loading configuration from file {config_file}')
        return _file_reader(config_file)
    elif s3:
        logger.debug(f'loading configuration from s3 {s3}')
        return _s3_reader(s3)
    elif conf_env:
        logger.debug(f'loading configuration from file {conf_env} via '
                'environment variable')
        return _file_reader(conf_env)
    elif s3_env:
        logger.debug(f'loading configuration from s3 {s3_env} via '
                'environment variable')
        return _s3_reader(s3_env)
    else:
        logger.debug('using default config file pydomotic.yml')
        return _file_reader('pydomotic.yml')

class _reader(object):
    def __init__(self, data):
        self.data = data

class _file_reader(_reader):
    def read(self):
        if not self.data:
            logger.warning('no config file or s3 data provided, skipping')
        elif not os.path.isfile(self.data):
            logger.warning(f'configuration file {self.data} does not exist, skipping')
        else:
            with open(self.data) as f:
                return f.read()

class _s3_reader(_reader):
    def parse(self):
        try:
            if isinstance(self.data, str):
                self.data = self.data.split('/', 1)
            bucket, key = self.data
            assert bucket and key
            return bucket, key
        except:
            raise PyDomoticConfigParsingError('malformed s3 object: '
                    'expecting tuple like (bucket, key) or string like "bucket/key"')

    def read(self):
        bucket, key = self.parse()

        try:
            import boto3
            client = boto3.client('s3')
            response = client.get_object(Bucket=bucket, Key=key)
            logger.debug('configuration successfully fetched from s3')
            return response['Body'].read()
        except Exception as e:
            logger.error('unable to fetch configuration from s3: '
                    f'[{e.__class__.__name__}] {e}')
            return None

def _parse_providers(providers_conf):
    providers = {
            'noop': NoopProvider(),
    }
    for name, provider in providers_conf.items():
        logging.info(f'preparing provider {name}')
        if name == 'tuya':
            providers['tuya'] = _parse_tuya_provider(provider)
        elif name == 'noop':
            pass  # already added
        elif name == 'fujitsu':
            providers['fujitsu'] = _parse_fujitsu_provider(provider)
        elif name == 'airthings':
            providers['airthings'] = _parse_airthings_provider(provider)
        elif name == 'moen':
            providers['moen'] = _parse_moen_provider(provider)
        elif name == 'ecobee':
            providers['ecobee'] = _parse_ecobee_provider(provider)
        else:
            raise PyDomoticConfigParsingError(f'unknown provider "{name}"')
    return providers

def _parse_tuya_provider(provider):
    for key in ('username', 'password', 'access_id', 'access_key'):
        if key not in provider:
            raise PyDomoticConfigParsingError(
                    f'provider tuya requires key "{key}"')

    username = _parse_string(provider['username'])
    password = _parse_string(provider['password'])
    access_id = _parse_string(provider['access_id'])
    access_key = _parse_string(provider['access_key'])

    cache_secs = provider.get('device_status_cache_seconds')
    timeout = provider.get('timeout_seconds')

    from .providers.tuya import TuyaProvider
    return TuyaProvider(username, password, access_id, access_key,
            status_cache_seconds=cache_secs, timeout=timeout)

def _parse_fujitsu_provider(provider):
    for key in ('username', 'password'):
        if key not in provider:
            raise PyDomoticConfigParsingError(
                    f'provider fujitsu requires key "{key}"')

    username = _parse_string(provider['username'])
    password = _parse_string(provider['password'])

    from .providers.fujitsu import FujitsuProvider
    return FujitsuProvider(username, password)

def _parse_airthings_provider(provider):
    for key in ('client_id', 'client_secret'):
        if key not in provider:
            raise PyDomoticConfigParsingError(
                    f'provider airthings requires key "{key}"')

    client_id = _parse_string(provider['client_id'])
    client_secret = _parse_string(provider['client_secret'])

    cache_secs = provider.get('data_cache_seconds')
    timeout = provider.get('timeout_seconds')

    from .providers.airthings import AirthingsProvider
    return AirthingsProvider(client_id, client_secret, data_cache_seconds=cache_secs,
            timeout=timeout)

def _parse_moen_provider(provider):
    for key in ('username', 'password'):
        if key not in provider:
            raise PyDomoticConfigParsingError(
                    f'provider moen requires key "{key}"')

    username = _parse_string(provider['username'])
    password = _parse_string(provider['password'])

    from .providers.moen import MoenProvider
    return MoenProvider(username, password)

def _parse_ecobee_provider(provider):
    for key in ('app_key', 'refresh_token'):
        if key not in provider:
            raise PyDomoticConfigParsingError(
                    f'provider ecobee requires key "{key}"')

    app_key = _parse_string(provider['app_key'])
    refresh_token = _parse_string(provider['refresh_token'])

    from .providers.ecobee import EcobeeProvider
    return EcobeeProvider(app_key, refresh_token)

_env_re = re.compile(r'\$\{env:(.*?)\}')
def _parse_string(string):
    def _replace_env(m):
        key = m.group(1)
        val = os.environ.get(m.group(1))
        if val is None:
            raise PyDomoticConfigParsingError(
                    f'expected environment variable "{key}" not found')
        return val
    return _env_re.sub(_replace_env, string)

def _parse_devices(devices_conf, providers):
    devices = {}
    for name, device in devices_conf.items():
        logger.info(f'preparing device {name}')
        provider_name = device.get('provider')
        if provider_name is None:
            raise PyDomoticConfigParsingError(
                    f'no provider given for device "{name}"')
        provider = providers.get(provider_name)
        if provider is None:
            raise PyDomoticConfigParsingError(
                    f'device "{name}" expected provider "{provider_name}" '
                    'not found')
        device_id = device.get('id')
        if device_id is None:
            raise PyDomoticConfigParsingError(
                    f'no id given for device "{name}"')
        try:
            description = device.get('description')
            devices[name] = provider.get_device(device_id, name, description)
        except Exception as e:
            raise PyDomoticConfigParsingError(
                    f'unable to get device "{name}": {e}')
    return devices

def _parse_components(automations, context):
    components = []
    for name, automation in automations.items():
        logger.info(f'preparing automation {name}')
        if not automation.get('enabled', True):
            continue
        for num, component in enumerate(automation.get('components', [])):
            component_name = f'{name} {num}'
            ifs = component.get('if') or {}
            thens = component.get('then') or {}
            elses = component.get('else') or {}
            logger.info(f'adding component {component_name}')
            components.append(Component(
                name=component_name,
                ifs=_parse_triggers(ifs, context),
                thens=_parse_actions(thens, context),
                elses=_parse_actions(elses, context),
                enabled=True,
            ))
    return components

def _parse_triggers(ifs, context):
    triggers = []
    for trigger_type, trigger_value in ifs.items():
        logger.debug(f'adding {trigger_type} trigger')
        triggers.append(_parse_trigger(trigger_type, trigger_value, context))
    return triggers

def _parse_trigger(typ, value, context, sensor=None):
    if typ == 'aqi':
        trigger = _parse_aqi_trigger(value, context, sensor=sensor)
    elif typ == 'time':
        trigger = _parse_time_trigger(value, context, sensor=sensor)
    elif typ == 'weekday':
        trigger = _parse_weekday_trigger(value, context, sensor=sensor)
    elif typ == 'date':
        trigger = _parse_date_trigger(value, context, sensor=sensor)
    elif typ == 'cron':
        trigger = _parse_cron_trigger(value, context, sensor=sensor)
    elif typ == 'random':
        trigger = _parse_random_trigger(value, context, sensor=sensor)
    elif typ == 'sunrise':
        trigger = _parse_sunrise_trigger(value, context, sensor=sensor)
    elif typ == 'sunset':
        trigger = _parse_sunset_trigger(value, context, sensor=sensor)
    elif typ == 'temp':
        trigger = _parse_temp_trigger(value, context, sensor=sensor)
    elif typ == 'radon':
        trigger = _parse_radon_trigger(value, context, sensor=sensor)
    elif typ == 'webhook':
        # TODO: test _parse_triggers
        trigger = _parse_webhook_trigger(value, context, sensor=sensor)
    elif typ in context.devices:
        if sensor:
            raise PyDomoticConfigParsingError('nested sensor confs not allowed')
        trigger = _parse_sensor_trigger(typ, value, context)
    else:
        raise PyDomoticConfigParsingError(f'unknown trigger type "{typ}"')
    return trigger

_ranged_value_aqi_re = re.compile(r'(<|>|==|<=|>=)?\s*(\d+\.?\d*)')
_ranged_value_temp_re = re.compile(r'(<|>|==|<=|>=)?\s*(\d+\.?\d*|\$\{temp\})')
def _parse_ranged_values(value, typ, context):
    if not isinstance(value, (str, int, float)):
        raise PyDomoticConfigParsingError(
                f'invalid {typ} trigger value "{value}", expecting string or '
                'number value like ">100", "<100", or "100"')

    def _ranged_func(start, end):
        return lambda a: a >= start and a <= end

    def _relative_func(op, val):
        value_func = _get_value_func(val)
        if op == '>':
            return lambda a: a > value_func()
        elif op == '<':
            return lambda a: a < value_func()
        elif op == '==':
            return lambda a: a == value_func()
        elif op == '>=':
            return lambda a: a >= value_func()
        elif op == '<=':
            return lambda a: a <= value_func()
        elif op == None:
            return lambda a: a == value_func()

    def _float_value_func(val):
        num = float(val)
        return lambda: num

    if typ == 'aqi' or typ == 'radon':
        _ranged_value_re = _ranged_value_aqi_re
        _get_value_func = lambda a: _float_value_func(a)
    elif typ == 'temp':
        _ranged_value_re = _ranged_value_temp_re
        def _get_value_func(val):
            if val == '${temp}':
                return context.weather_sensor.current_temperature
            return _float_value_func(val)
    else:
        raise PyDomoticConfigParsingError(f'unknown ranged trigger type "{typ}"')

    _check_funcs = []
    for val in str(value).split(','):
        val = val.strip().lower()
        ranged_val = val.split('-')

        if len(ranged_val) == 1:
            m = _ranged_value_re.fullmatch(ranged_val[0].strip())
            if not m:
                raise PyDomoticConfigParsingError(
                        f'invalid {typ} trigger value "{val}", expecting value like '
                        '">100", "<100", or "100"')
            _check_func =_relative_func(m.group(1), m.group(2))

        elif len(ranged_val) == 2:
            try:
                start_val = float(ranged_val[0])
                end_val = float(ranged_val[1])
                _check_func = _ranged_func(start_val, end_val)
            except:
                raise PyDomoticConfigParsingError(
                        f'invalid {typ} trigger value "{val}", expecting ranged '
                        'value like "80-100"')

        else:
            raise PyDomoticConfigParsingError(f'unknown {typ} "{val}"')

        _check_funcs.append(_check_func)
    return lambda a: any(fn(a) for fn in _check_funcs)

def _parse_aqi_trigger(value, context, sensor=None):
    _check_func = _parse_ranged_values(value, 'aqi', context)
    sensor = sensor or context.aqi_sensor
    return AQITrigger(_check_func, sensor)

_time_re = re.compile(r'(10|11|12|[1-9]):([0-5][0-9])\s*([ap]m)')
def _parse_time_trigger(value, context, sensor=None):
    def _parse_single_time(time):
        m = _time_re.fullmatch(time)
        if not m:
            raise PyDomoticConfigParsingError(
                    f'unknown time "{time}", expecting time like "HH:MMam"')
        hour, minute = int(m.group(1)), int(m.group(2))
        if hour == 12:
            hour = 0
        if m.group(3) == 'pm':
            hour += 12
        return 60*hour + minute

    times = []
    for time in value.split(','):
        time = time.strip().lower()
        ranged_time = time.split('-')
        start_time = _parse_single_time(ranged_time[0].strip())
        if len(ranged_time) == 1:
            end_time = start_time + 1
        elif len(ranged_time) == 2:
            end_time = _parse_single_time(ranged_time[1].strip()) + 1
        else:
            raise PyDomoticConfigParsingError(
                    f'unknown time "{time}", expecting time like '
                    '"HH:MMam-HH:MMam"')
        for minute in range(start_time, end_time):
            times.append(minute)
    return TimeTrigger(times, time_sensor=sensor or context.time_sensor)

_isoweekdays = {
        'monday': 1,
        'mon': 1,
        'tuesday': 2,
        'tues': 2,
        'tue': 2,
        'wednesday': 3,
        'wed': 3,
        'thursday': 4,
        'thurs': 4,
        'thur': 4,
        'thu': 4,
        'friday': 5,
        'fri': 5,
        'saturday': 6,
        'sat': 6,
        'sunday': 7,
        'sun': 7,
}

def _parse_weekday_trigger(value, context, sensor=None):
    def _parse_single_day(weekday):
        isoweekday = _isoweekdays.get(weekday)
        if isoweekday is None:
            raise PyDomoticConfigParsingError(
                    f'"{weekday}" is not a valid weekday')
        return isoweekday

    isoweekdays = []
    for weekday in value.split(','):
        weekday = weekday.strip().lower()
        ranged_weekday = weekday.split('-')
        start_weekday = _parse_single_day(ranged_weekday[0].strip())
        if len(ranged_weekday) == 1:
            end_weekday = start_weekday + 1
        elif len(ranged_weekday) == 2:
            end_weekday = _parse_single_day(ranged_weekday[1].strip()) + 1
            if end_weekday <= start_weekday:
                end_weekday += 7
        else:
            raise PyDomoticConfigParsingError(f'unknown weekday "{weekday}"')
        for isoweekday in range(start_weekday, end_weekday):
            isoweekdays.append(isoweekday % 7 or 7)
    return IsoWeekdayTrigger(isoweekdays, time_sensor=sensor or context.time_sensor)

def _parse_date_trigger(value, context, sensor=None):
    if isinstance(value, datetime.date):
        dates = [value]
    else:
        dates = []
        for date in value.split(','):
            try:
                date = datetime.date.fromisoformat(date.strip().lower())
            except:
                raise PyDomoticConfigParsingError(
                        f'unknown date "{date}", expecting date like "YYYY-MM-DD"')
            dates.append(date)
    return DateTrigger(dates, time_sensor=sensor or context.time_sensor)

def _parse_cron_trigger(value, context, sensor=None):
    if not croniter.croniter.is_valid(value):
        raise PyDomoticConfigParsingError(f'invalid cron expression "{value}"')
    return CronTrigger(value, time_sensor=sensor or context.time_sensor)

def _parse_random_trigger(value, context, sensor=None):
    # TODO: check for valid probability
    probability = float(value)
    return RandomTrigger(probability)

_ranged_timedelta_re = re.compile(r'(-?\d+)((\s*-\s*)(-?\d+))?')
def _parse_timedelta(value):
    timedeltas = []
    for delta in str(value).split(','):
        m = _ranged_timedelta_re.fullmatch(delta.strip())
        if not m:
            raise PyDomoticConfigParsingError(
                    f'unknown time delta "{value}", expecting value like '
                    '"60", "60-120", or "2:15"')
        start_delta = int(float(m.group(1)))
        end_delta = int(float(m.group(4) or start_delta))
        for i in range(start_delta, end_delta+1):
            timedeltas.append(i)
    return timedeltas

def _parse_sunrise_trigger(value, context, sensor=None):
    timedelta = _parse_timedelta(value)
    sun_sensor = sensor or context.sun_sensor
    return SunriseTrigger(timedelta, context.time_sensor, sun_sensor)

def _parse_sunset_trigger(value, context, sensor=None):
    timedelta = _parse_timedelta(value)
    sun_sensor = sensor or context.sun_sensor
    return SunsetTrigger(timedelta, context.time_sensor, sun_sensor)

def _parse_temp_trigger(value, context, sensor=None):
    _check_func = _parse_ranged_values(value, 'temp', context)
    # TODO: test weather sensor singleton
    sensor = sensor or context.weather_sensor
    return TemperatureTrigger(_check_func, sensor)

def _parse_radon_trigger(value, context, sensor=None):
    if not sensor:
        raise PyDomoticConfigParsingError(
                'radon trigger requires a radon sensor')
    _check_func = _parse_ranged_values(value, 'radon', context)
    return RadonTrigger(_check_func, sensor)

def _parse_webhook_trigger(value, context, sensor=None):
    # TODO: test _parse_webhook_trigger
    return WebhookTrigger(value, sensor or context.webhook_sensor)

def _parse_sensor_trigger(device_name, value, context):
    # TODO: test value is None
    if value is None:
        raise PyDomoticConfigParsingError(
                f'sensor trigger for {device_name} requires a value')
    sensor = context.device_sensor(device_name)
    trigger_type, trigger_value = value.popitem()
    if value:
        raise PyDomoticConfigParsingError(
                f'sensor trigger {value} ignored, only one trigger allowed '
                'per sensor config')
    return _parse_trigger(trigger_type, trigger_value, context=context,
            sensor=sensor)

def _parse_actions(thens, context):
    actions = []
    for action_type, action_value in thens.items():
        logger.debug(f'adding {action_type} action')
        if action_type == 'exec':
            for method_name in action_value.split(','):
                method_name = method_name.strip()
                actions.append(ExecuteCodeAction(method_name, context.context))
        elif action_type == 'set-mode':
            actions.append(_parse_set_mode_action(action_value, context))
        else:
            if action_type == 'turn-on':
                cls = TurnOnAction
            elif action_type == 'turn-off':
                cls = TurnOffAction
            elif action_type == 'switch':
                cls = SwitchAction
            else:
                raise PyDomoticConfigParsingError(
                        f'unknown action type "{action_type}"')
            for device_name in action_value.split(','):
                device_name = device_name.strip()
                device = context.devices.get(device_name)
                if device is None:
                    raise PyDomoticConfigParsingError(
                            f'unknown device name "{device_name}"')
                actions.append(cls(device))
    return actions

def _parse_set_mode_action(value, context):
    if not isinstance(value, dict):
        raise PyDomoticConfigParsingError(
                f'set-mode action requires a dict with keys "device" and "mode"')

    for key in ('device', 'mode'):
        if key not in value:
            raise PyDomoticConfigParsingError(
                    f'set-mode action requires key "{key}"')

    device_name = value['device']
    device = context.devices.get(device_name)
    if device is None:
        raise PyDomoticConfigParsingError(
                f'unknown device name "{device_name}"')

    mode = value['mode']
    if mode not in ('home', 'away', 'sleep'):
        raise PyDomoticConfigParsingError(
                f'unknown mode "{mode}", expecting "home", "away", or "sleep"')

    extra_params = {}
    if mode == "sleep":
        if 'revert-min' in value:
            extra_params['revertMinutes'] = int(value['revert-min'])
        if 'revert-mode' in value:
            rmode = value['revert-mode']
            if rmode not in ('home', 'away'):
                raise PyDomoticConfigParsingError(
                        f'unknown mode "{rmode}", expecting "home" or "away"')
            extra_params['revertMode'] = value['revert-mode']

    return SetModeAction(device, mode, extra_params)



import gosundpy
import logging

from .base import Provider, Device

logger = logging.getLogger(__name__)

class TuyaProvider(Provider):

    def __init__(self, username, password, access_id, access_key,
            status_cache_seconds=None, timeout=None):
        self.tuya = gosundpy.Gosund(username, password, access_id, access_key,
                status_cache_seconds=status_cache_seconds, timeout=timeout)

    def get_device(self, device_id, device_name, device_description):
        device = self.tuya.get_device(device_id)
        return TuyaDevice(device, device_name, device_description)

class TuyaDevice(Device):

    def turn_on(self):
        self.device.turn_on()
        logger.debug('device "%s" turned on', self.name)

    def turn_off(self):
        self.device.turn_off()
        logger.debug('device "%s" turned off', self.name)

    def switch(self):
        self.device.switch()
        logger.debug('device "%s" switched', self.name)

    def current_temperature(self):
        return self.device.get_temperature()



import abc
import functools
import importlib
import logging
import re
import time

from .exceptions import PyDomoticMethodImportError

logger = logging.getLogger(__name__)

class _timed_cache(object):

    def __init__(self):
        self.reset()

    def set(self, last_call, value):
        self.last_call = last_call
        self.value = value

    def reset(self):
        self.last_call = 0
        self.value = None

def cache_value(hours=0, minutes=0, seconds=0, fallback_on_error=False):
    # XXX: cache stored per class rather than per instance
    seconds += 60 * 60 * hours + 60 * minutes
    cache = _timed_cache()
    def _rate_limit(fn):
        @functools.wraps(fn)
        def _call(*args, **kwargs):
            now = time.time()
            if now - cache.last_call < seconds:
                return cache.value
            try:
                cache.set(now, fn(*args, **kwargs))
            except Exception as e:
                if not fallback_on_error or cache.last_call == 0:
                    raise e
                logger.info(f'falling back to cached value: [{e.__class__.__name__}] {e}')
            return cache.value
        _call.clear_cache = cache.reset
        return _call
    return _rate_limit

_camel_to_snake_re_1 = re.compile('(.)([A-Z][a-z]+)')
_camel_to_snake_re_2 = re.compile('([a-z0-9])([A-Z])')
def _camel_to_snake(name):
    name = _camel_to_snake_re_1.sub(r'\1_\2', name)
    return _camel_to_snake_re_2.sub(r'\1_\2', name).lower()

class _Nameable(object):

    @property
    def name(self):
        if not hasattr(self, '_name'):
            self._name = _camel_to_snake(self.__class__.__name__)
        return self._name

class ObjectMetaclass(abc.ABCMeta):

    extra_bases = (_Nameable,)

    def __new__(cls, clsname, bases, dct):
        return super().__new__(cls, clsname, bases + cls.extra_bases, dct)

    def __call__(cls, *args, **kwargs):
        # acts as __init__ for any objects using this metaclass
        obj = super().__call__(*args, **kwargs)
        obj.required_class_attrs = getattr(cls, 'required_class_attrs', [])
        for attr in obj.required_class_attrs:
            if not hasattr(obj, attr):
                raise TypeError(
                        "Can't instantiate abstract class "
                        f'{cls.__name__} with abstract instance '
                        f'variable {attr}')
        return obj

def import_method(import_path):
    names = import_path.rsplit('.', 1)
    if len(names) == 1:
        raise PyDomoticMethodImportError(
                f'import path must include module and method name')
    package_name, method_name = names
    try:
        package = importlib.import_module(package_name)
        return getattr(package, method_name)
    except Exception as e:
        raise PyDomoticMethodImportError(
                f'unable to import code "{import_path}": '
                f'[{e.__class__.__name__}] {e}')



from .exceptions import (PyDomoticConfigParsingError,
        PyDomoticComponentRunError, PyDomoticMethodImportError)
from .handlers import Handler, LambdaHandler
from .version import version

__all__ = [
        'Handler',
        'LambdaHandler',
        'PyDomoticConfigParsingError',
        'PyDomoticComponentRunError',
        'PyDomoticMethodImportError',
]
__version__ = version



import abc

from ..utils import ObjectMetaclass

class Provider(metaclass=ObjectMetaclass):

    @abc.abstractmethod
    def get_device(self, device_id, device_name, device_description):
        pass

class Device(metaclass=ObjectMetaclass):

    def __init__(self, device, name, description):
        self.device = device
        self.device_name = name
        self.device_description = description

    @property
    def name(self):
        return f'{super().name} {self.device_name}'



import logging
import requests
import time

from pydomotic.providers.base import Provider, Device

logger = logging.getLogger(__name__)

class EcobeeProvider(Provider):

    def __init__(self, app_key, refresh_token):
        self.api = EcobeeAPI(app_key, refresh_token)

    def get_device(self, device_id, device_name, device_description):
        device = self.api.get_device(device_id)
        return EcobeeDevice(device, device_name, device_description)

class EcobeeDevice(Device):

    def turn_on(self):
        self.device.turn_fan_on()
        logger.debug('device "%s" turned on', self.name)

    def turn_off(self):
        self.device.turn_fan_off()
        logger.debug('device "%s" turned off', self.name)

    def current_temperature(self):
        return self.device.get_temperature()

class EcobeeAPI(object):

    api_url = 'https://api.ecobee.com'
    token_url = api_url + '/token'
    thermostat_url = api_url + '/1/thermostat'

    headers = {
        'Content-Type': 'application/json;charset=UTF-8',
    }

    def __init__(self, app_key, refresh_token):
        self._refresh_token = refresh_token
        self._app_key = app_key
        self._expires_at = 0

    def get_device(self, device_id):
        return self.device(device_id, self)

    def get_thermostat(self):
        return self._make_request('GET', self.thermostat_url, params={
            'format': 'json',
            'body': """{
                "selection": {
                    "selectionType": "registered",
                    "selectionMatch": "",
                    "includeSensors": true
                }
            }""",
        })

    def get_sensor_data(self, sensor_id):
        for thermostat in self.get_thermostat()['thermostatList']:
            for sensor in thermostat['remoteSensors']:
                if sensor['id'] == sensor_id:
                    return sensor
        raise EcobeeError(f'sensor "{sensor_id}" not found')

    def set_fan_hold(self, mode):
        self._make_request('POST', self.thermostat_url, params={
            'format': 'json',
            'body': f"""{{
                "selection": {{
                    "selectionType":"registered",
                    "selectionMatch":""
                }},
                "functions": [
                    {{
                        "type":"setHold",
                        "params":{{
                            "holdType":"indefinite",
                            "fan":"{mode}"
                        }}
                    }}
                ]
            }}""",
        })

    def _authenticate(self):
        resp = self._make_request('POST', self.token_url, params={
            'grant_type': 'refresh_token',
            'client_id': self._app_key,
            'code': self._refresh_token,
        })
        self._expires_at = time.time() + resp.get('expires_in')
        access_token, token_type = resp.get('access_token'), resp.get('token_type')
        self.headers['Authorization'] = f'{token_type} {access_token}'

    def _make_request(self, method, url, params=None):
        if time.time() > self._expires_at and url != self.token_url:
            self._authenticate()
        resp = requests.request(method, url, params=params, headers=self.headers)
        resp.raise_for_status()
        return resp.json()

    class device(object):

        def __init__(self, device_id, api):
            self.device_id = device_id
            self.api = api

        def turn_fan_on(self):
            self.api.set_fan_hold('on')

        def turn_fan_off(self):
            self.api.set_fan_hold('auto')

        def get_temperature(self):
            data = self.api.get_sensor_data(self.device_id)
            for cap in data['capability']:
                if cap['type'] == 'temperature':
                    return int(cap['value']) / 10
            raise EcobeeError(f'temperature sensor not found for device "{self.device_id}"')

        def get_humidity(self):
            data = self.api.get_sensor_data(self.device_id)
            for cap in data['capability']:
                if cap['type'] == 'humidity':
                    return int(cap['value'])
            raise EcobeeError(f'humidity sensor not found for device "{self.device_id}"')

class EcobeeError(Exception):
    pass



import codecs
import os.path

from setuptools import find_packages, setup

def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()

def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('version'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError('Unable to find version string.')

def get_extras_require():
    # providers
    airthings = []
    ecobee = []
    fujitsu = ['pyfujitsu>=0.9.0,<1.0']
    moen = ['pyflowater>=0.5.2,<1.0', 'retry>=0.9.2,<1.0']
    tuya = ['gosundpy>=0.7.0,<1.0']

    # features
    s3 = ['boto3>=1.26.3,<2.0']
    tz = ['timezonefinder>=6.1.8,<7.0']

    return {
        'airthings': airthings,
        'ecobee': ecobee,
        'fujitsu': fujitsu,
        'moen': moen,
        'tuya': tuya,
        's3': s3,
        'tz': tz,
        'all': airthings + ecobee + fujitsu + moen + tuya + s3 + tz,
    }

setup(
    name='pydomotic',
    version=get_version('pydomotic/version.py'),
    url='https://github.com/purple4reina/pydomotic',
    author='Rey Abolofia',
    author_email='purple4reina@gmail.com',
    keywords='iot,python,home,automation,smart,domotic',
    description=('Python library for home automation execution, enabling '
            'seamless control and management of your IoT devices'),
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    license='MIT',
    project_urls={
        'Source': 'https://github.com/purple4reina/pydomotic',
        'Bug Tracker': 'https://github.com/purple4reina/pydomotic/issues',
    },
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Software Development',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    install_requires=[
        'PyYAML>=6.0,<7.0',
        'astral>=3.0,<4.0',
        'croniter>=1.3.15,<2.0',
        'pyowm>=3.3.0,<4.0',
    ],
    extras_require=get_extras_require(),
    packages=find_packages(),
    python_requires='>=3.8,<4',
)



import logging

from .base import Provider, Device

logger = logging.getLogger(__name__)

class NoopProvider(Provider):

    def get_device(self, device_id, device_name, device_description):
        return NoopDevice(device_id, device_name, device_description)

class NoopDevice(Device):

    def turn_on(self):
        logger.debug('device "%s" turned on', self.name)

    def turn_off(self):
        logger.debug('device "%s" turned off', self.name)

    def switch(self):
        logger.debug('device "%s" switched', self.name)

    def current_temperature(self):
        return 42



import logging
import time
import traceback

from .exceptions import PyDomoticComponentRunError
from .parsers import parse_yaml

logger = logging.getLogger(__name__)

class Handler(object):

    def __init__(self, config_file=None):
        self.components, _ = parse_yaml(config_file=config_file)

    def __call__(self):
        self.run_components()

    def run_components(self):
        components, failed = [c for c in self.components if c.enabled], []
        attempts = 3

        while components and attempts:
            attempts -= 1
            for component in components:
                try:
                    component.run()
                except Exception:
                    exc = ''.join(traceback.format_exc())
                    logger.error(
                        f'failure running component {component.name}, '
                        f'{attempts} remaining attempts\n{exc}')
                    failed.append(component)

            if failed and attempts:
                time.sleep(0.25)
            components, failed = failed, []

        if components:
            raise PyDomoticComponentRunError(
                    f'one or more components failed after 3 attempts: '
                    f'{", ".join(c.name for c in components)}')

class LambdaHandler(Handler):

    ok_response = {
            'statusCode': 200,
            'body': '{"status":"ok"}',
    }

    def __init__(self, config_file=None, s3=None):
        self.components, context = parse_yaml(config_file=config_file, s3=s3)
        self.webhook_sensor = context.webhook_sensor

    def __call__(self, event, context):
        # TODO: test webhook triggers
        self.webhook_sensor.set_webhook_request(event)
        self.run_components()
        return self.ok_response

class CommandLineHandler(Handler):

    def __init__(self):
        args = self.parse_args()
        super().__init__(config_file=args.config_file)

    def parse_args(self):
        import argparse
        parser = argparse.ArgumentParser(
                prog='python -m pydomotic',
                description='run pydomotic components',
        )
        parser.add_argument(
                '-c', '--config-file',
                help=('path to config file, will default to pydomotic.yaml '
                        'if no other config setting is found'),
        )
        return parser.parse_args()
